#include "Best.hpp"
#include "StringUtils.hpp"
#include "AtomicFileUtils.hpp"
#include "CalcScore.hpp"

using std::string;
using std::vector;

Best::Best() :
    m_id(0),
    m_score(0),
    m_similarity(0),
    m_cost(0)
{
}

Best::~Best()
{
}

namespace
{
    string getFileName(uint64_t id)
    {
        return strprintf("../solutions/%" PRIu64 "", id);
    }
}


void Best::load(uint64_t id)
{
    m_id = id;

    string msg;
    if (!loadProblem(id, &m_problem, &msg))
    {
        fprintf(stderr, "%s\n", msg.c_str());
        exit(1);
    }

    m_solutionData.clear();

    string fileName = getFileName(id);
    atomicReadFile(fileName, &m_solutionData);

    parse();
}

bool Best::update(const Solution& solution)
{
    uint64_t id = m_id;

    if (id == 0)
    {
        return false;
    }

    Image result;
    uint64_t score;
    uint64_t similarity;
    uint64_t cost;
    if (!calcScore(m_problem, solution, &result, &score, &similarity, &cost))
    {
        return false;
    }

    while (true)
    {
        if (score >= m_score)
        {
            return false;
        }

        m_solution = solution;
        m_score = score;
        m_similarity = similarity;
        m_cost = cost;

        string solutionData = formatSolution(m_problem, m_solution);

        string fileName = getFileName(id);

        if (atomicReplaceFile(fileName, &m_solutionData, solutionData))
        {
            m_solutionData.swap(solutionData);
            return true;
        }

        parse();
    }
}

void Best::parse()
{
    string msg("parse");

    if (!parseSolution(m_problem, m_solutionData, &m_solution) ||
        !calcScore(m_problem, m_solution, &m_result, &m_score, &m_similarity, &m_cost, nullptr, &msg))
    {
        m_solution.clear();
        if (!calcScore(m_problem, m_solution, &m_result, &m_score, &m_similarity, &m_cost))
        {
            fprintf(stderr, "Evaluating empty solution failed\n");
            exit(1);
        }
    }
}
