#include "CalcSimilarity.hpp"
#include <cmath>

namespace
{
    double pixelDiff(Color p1, Color p2)
    {
        uint32_t r = (uint32_t)p1.r - (uint32_t)p2.r;
        uint32_t g = (uint32_t)p1.g - (uint32_t)p2.g;
        uint32_t b = (uint32_t)p1.b - (uint32_t)p2.b;
        uint32_t a = (uint32_t)p1.a - (uint32_t)p2.a;

        r *= r;
        g *= g;
        b *= b;
        a *= a;

        return std::sqrt(r + g + b + a);
    }
}

uint64_t calcSimilarity(const Image& target,
                        const Image& result,
                        Image* pErrorImage)
{
    size_t width = target.getWidth();
    size_t height = target.getHeight();
    if (result.getWidth() != width ||
        result.getHeight() != height)
    {
        fprintf(stderr, "calcSimilarity: size mismatch");
        exit(1);
    }

    double diff = 0.0;

    if (pErrorImage)
    {
        pErrorImage->resize(width, height);
        for (size_t y = 0; y < height; y++)
        {
            for (size_t x = 0; x < width; x++)
            {
                double curDiff = pixelDiff(target(x, y), result(x, y));
                diff += curDiff;
                uint32_t intError = (uint32_t)(curDiff * (255.0 / 442.0));
                (*pErrorImage)(x, y) = Color(intError, 0, 0, 255);
            }
        }
    }
    else
    {
        for (size_t y = 0; y < height; y++)
        {
            for (size_t x = 0; x < width; x++)
            {
                diff += pixelDiff(target(x, y), result(x, y));
            }
        }
    }

    double alpha = 0.005;
    return (uint64_t)(diff * alpha + 0.5);
}
