#ifndef FASTRANDOM_HPP
#define FASTRANDOM_HPP

#include "Common.hpp"
#include <random>

// fastRandom
//
// Generate a random number from 0 to range-1 without using division.
// Use as a faster alternative to std::uniform_int_distribution.
// Assumes that Gen returns a full 32 or 64 bits of random data.

template<typename Gen, typename = std::enable_if_t<std::is_same_v<typename Gen::result_type, uint64_t>>>
uint64_t fastRandom(uint64_t range, Gen& gen)
{
    return ((__int128)range * gen()) >> 64;
}

template<typename Gen, typename = std::enable_if_t<std::is_same_v<typename Gen::result_type, uint32_t>>>
uint32_t fastRandom(uint32_t range, Gen& gen)
{
    return ((uint64_t)range * gen()) >> 32;
}

// fastShuffle
//
// Shuffle the given range into a random order without using division.
// Use as a faster alternative to std::shuffle.
// Assumes that Gen returns a full 32 or 64 bits of random data.

template<class It, class Gen>
void fastShuffle(It first, It last, Gen& gen)
{
    size_t size = last - first;
    while (size > 1)
    {
        size_t pos = fastRandom(size, gen);
        size--;
        using std::swap;
        swap(*(first + pos), *(first + size));
    }
}

#endif
