#ifndef JSONUTILS_HPP
#define JSONUTILS_HPP

#include "Common.hpp"
#include "Json.hpp"

class JException
{
public:
    JException(const char* msg) : m_msg(msg) {}

    const char* m_msg;
};

inline bool JGetBool_(const rapidjson::Value& jValue, const char* msg)
{ if (!jValue.IsBool()) throw JException(msg); return jValue.GetBool(); }
inline int32_t JGetInt_(const rapidjson::Value& jValue, const char* msg)
{ if (!jValue.IsInt()) throw JException(msg); return jValue.GetInt(); }
inline uint32_t JGetUint_(const rapidjson::Value& jValue, const char* msg)
{ if (!jValue.IsUint()) throw JException(msg); return jValue.GetUint(); }
inline int64_t JGetInt64_(const rapidjson::Value& jValue, const char* msg)
{ if (!jValue.IsInt64()) throw JException(msg); return jValue.GetInt64(); }
inline uint64_t JGetUint64_(const rapidjson::Value& jValue, const char* msg)
{ if (!jValue.IsUint64()) throw JException(msg); return jValue.GetUint64(); }
inline double JGetDouble_(const rapidjson::Value& jValue, const char* msg)
{ if (!jValue.IsNumber()) throw JException(msg); return jValue.GetDouble(); }
inline const char* JGetString_(const rapidjson::Value& jValue, const char* msg)
{ if (!jValue.IsString()) throw JException(msg); return jValue.GetString(); }
inline const rapidjson::Value& JCheckObject_(const rapidjson::Value& jValue,
                                             const char* msg)
{ if (!jValue.IsObject()) throw JException(msg); return jValue; }
inline const rapidjson::Value& JCheckArray_(const rapidjson::Value& jValue,
                                            const char* msg)
{ if (!jValue.IsArray()) throw JException(msg); return jValue; }

#define JGetBool(jValue) JGetBool_(jValue, #jValue ": not bool")
#define JGetInt(jValue) JGetInt_(jValue, #jValue ": not int")
#define JGetUint(jValue) JGetUint_(jValue, #jValue ": not uint")
#define JGetInt64(jValue) JGetInt64_(jValue, #jValue ": not int64")
#define JGetUint64(jValue) JGetUint64_(jValue, #jValue ": not uint64")
#define JGetDouble(jValue) JGetDouble_(jValue, #jValue ": not double")
#define JGetString(jValue) JGetString_(jValue, #jValue ": not string")
#define JCheckObject(jValue) JCheckObject_(jValue, #jValue ": not object")
#define JCheckArray(jValue) JCheckArray_(jValue, #jValue ": not array")

inline const rapidjson::Document& JCheckParse_(const rapidjson::Document& jDocument,
                                               const char* msg)
{ if (jDocument.HasParseError()) throw JException(msg); return jDocument; }

#define JCheckParse(jDocument) \
    JCheckParse_(jDocument, #jDocument ": parse failed")

#define JCheckParseObject(jDocument)                                  \
    JCheckObject_(JCheckParse(jDocument), #jDocument ": not object")
#define JCheckParseArray(jDocument) \
    JCheckArray_(JCheckParse(jDocument), #jDocument ": not array")

inline const rapidjson::Value& JObjectGet_(const rapidjson::Value& jObject,
                                           const char* key,
                                           const char* msg)
{
    auto jit = jObject.FindMember(key);
    if (jit == jObject.MemberEnd()) throw JException(msg);
    return jit->value;
}

#define JObjectGet(jObject, key) \
    JObjectGet_(jObject, key, #jObject ": '" key "' not found")

#define JObjectGetBool(jObject, key) \
    JGetBool_(JObjectGet(jObject, key), #jObject ": '" key "' not bool")
#define JObjectGetInt(jObject, key) \
    JGetInt_(JObjectGet(jObject, key), #jObject ": '" key "' not int")
#define JObjectGetUint(jObject, key) \
    JGetUint_(JObjectGet(jObject, key), #jObject ": '" key "' not uint")
#define JObjectGetInt64(jObject, key) \
    JGetInt64_(JObjectGet(jObject, key), #jObject ": '" key "' not int64")
#define JObjectGetUint64(jObject, key) \
    JGetUint64_(JObjectGet(jObject, key), #jObject ": '" key "' not uint64")
#define JObjectGetDouble(jObject, key) \
    JGetDouble_(JObjectGet(jObject, key), #jObject ": '" key "' not double")
#define JObjectGetString(jObject, key) \
    JGetString_(JObjectGet(jObject, key), #jObject ": '" key "' not string")
#define JObjectGetObject(jObject, key) \
    JCheckObject_(JObjectGet(jObject, key), #jObject ": '" key "' not object")
#define JObjectGetArray(jObject, key) \
    JCheckArray_(JObjectGet(jObject, key), #jObject ": '" key "' not array")

#endif
