#include "Merge.hpp"
#include "StringUtils.hpp"

using std::string;

bool merge(const Problem& problem,
           Canvas& canvas,
           Solution& solution)
{
    auto& blocks = canvas.blocks;

    uint64_t x = 0;

    while (x < problem.width)
    {
        bool foundBlock0 = false;
        uint64_t blockNum0 = 0;
        Rect rect0;

        for (uint64_t blockNum = 0; blockNum < blocks.size(); blockNum++)
        {
            auto& block = blocks[blockNum];
            if (!block.valid) continue;
            if (block.rect.lowX == x &&
                block.rect.lowY == 0)
            {
                foundBlock0 = true;
                blockNum0 = blockNum;
                rect0 = block.rect;
                break;
            }
        }

        if (!foundBlock0)
        {
            printf("1\n");
            return false;
        }

        if (rect0.highY == problem.height)
        {
            x = rect0.highX;
            continue;
        }

        bool foundBlock1 = false;
        uint64_t blockNum1 = 0;
        Rect rect1;

        for (uint64_t blockNum = 0; blockNum < blocks.size(); blockNum++)
        {
            auto& block = blocks[blockNum];
            if (!block.valid) continue;
            if (block.rect.lowX == x &&
                block.rect.highX == rect0.highX &&
                block.rect.lowY == rect0.highY)
            {
                foundBlock1 = true;
                blockNum1 = blockNum;
                rect1 = block.rect;
                break;
            }
        }

        if (!foundBlock1)
        {
            printf("2\n");
            return false;
        }

        canvas.blocks[blockNum0].valid = false;
        canvas.blocks[blockNum1].valid = false;

        {
            auto& move = solution.emplace_back();
            move.moveKind = MoveKind::Merge;
            move.blockNums[0] = blockNum0;
            move.blockNums[1] = blockNum1;
        }

        Rect mergeRect(rect0.lowX, rect0.lowY, rect0.highX, rect1.highY);

        {
            auto& block = canvas.blocks.emplace_back();
            block.rect = mergeRect;
            block.blockKind = BlockKind::Simple;
            block.blockId = canvas.nextBlockId++;
        }
    }

    while (true)
    {
        bool foundBlock0 = false;
        uint64_t blockNum0 = 0;
        Rect rect0;

        for (uint64_t blockNum = 0; blockNum < blocks.size(); blockNum++)
        {
            auto& block = blocks[blockNum];
            if (!block.valid) continue;
            if (block.rect.lowX == 0 &&
                block.rect.lowY == 0 &&
                block.rect.highY == problem.height)
            {
                foundBlock0 = true;
                blockNum0 = blockNum;
                rect0 = block.rect;
                break;
            }
        }

        if (!foundBlock0)
        {
            printf("3\n");
            return false;
        }

        if (rect0.highX == problem.width)
        {
            break;
        }

        bool foundBlock1 = false;
        uint64_t blockNum1 = 0;
        Rect rect1;

        for (uint64_t blockNum = 0; blockNum < blocks.size(); blockNum++)
        {
            auto& block = blocks[blockNum];
            if (!block.valid) continue;
            if (block.rect.lowX == rect0.highX &&
                block.rect.lowY == 0 &&
                block.rect.highY == problem.height)
            {
                foundBlock1 = true;
                blockNum1 = blockNum;
                rect1 = block.rect;
                break;
            }
        }

        if (!foundBlock1)
        {
            printf("4\n");
            return false;
        }

        canvas.blocks[blockNum0].valid = false;
        canvas.blocks[blockNum1].valid = false;

        {
            auto& move = solution.emplace_back();
            move.moveKind = MoveKind::Merge;
            move.blockNums[0] = blockNum0;
            move.blockNums[1] = blockNum1;
        }

        Rect mergeRect(rect0.lowX, rect0.lowY, rect1.highX, rect0.highY);

        {
            auto& block = canvas.blocks.emplace_back();
            block.rect = mergeRect;
            block.blockKind = BlockKind::Simple;
            block.blockId = canvas.nextBlockId++;
        }
    }

    uint64_t validCount = 0;
    for (uint64_t blockNum = 0; blockNum < blocks.size(); blockNum++)
    {
        auto& block = blocks[blockNum];
        if (!block.valid) continue;
        validCount++;
    }
    if (validCount != 1)
    {
        printf("5\n");
        return false;
    }

    return true;
}
