#include "Protocol.hpp"
#include "Http.hpp"
#include "FileUtils.hpp"
#include "StringUtils.hpp"

using std::string;
using std::vector;

namespace Protocol
{
    void init()
    {
        Http::init();
    }

    void cleanup()
    {
        Http::cleanup();
    }

    bool getAPIKey(string* pKey)
    {
        string tmpKey;
        if (!readFile("../../../apikey.txt", &tmpKey))
        {
            return false;
        }

        string key;
        for (char ch : tmpKey)
        {
            if (ch >= 33 && ch <= 126)
            {
                key.push_back(ch);
            }
        }
        if (key.empty())
        {
            return false;
        }

        if (pKey) *pKey = std::move(key);
        return true;
    }

    bool getAPIKeyHeader(vector<string>* pHeaders,
                         std::string* pMsg = nullptr)
    {
        string key;
        if (!getAPIKey(&key))
        {
            if (pMsg) *pMsg = "Error reading API key";
            return false;
        }

        pHeaders->push_back("Authorization: Bearer " + key);
        return true;
    }

    void getUserAgentHeader(vector<string>* pHeaders)
    {
        pHeaders->push_back("User-Agent: libcurl (Frictionless Bananas ICFPC 2022 tools)");
    }

    bool getSubmissions(string* pSubmissionsData,
                        string* pMsg)
    {
        vector<string> headers;
        if (!getAPIKeyHeader(&headers, pMsg)) return false;

        getUserAgentHeader(&headers);

        string url = "https://robovinci.xyz/api/submissions";

        string response;
        if (!Http::get(url, &headers, &response, pMsg))
        {
            return false;
        }

        if (pSubmissionsData) *pSubmissionsData = std::move(response);
        return true;
    }

    bool getResults(string* pResultsData,
                    string* pMsg)
    {
        vector<string> headers;
        if (!getAPIKeyHeader(&headers, pMsg)) return false;

        getUserAgentHeader(&headers);

        string url = "https://robovinci.xyz/api/results/user";

        string response;
        if (!Http::get(url, &headers, &response, pMsg))
        {
            return false;
        }

        if (pResultsData) *pResultsData = std::move(response);
        return true;
    }

    bool submit(uint64_t id,
                const string& solutionData,
                string* pMsg)
    {
        vector<string> headers;
        if (!getAPIKeyHeader(&headers, pMsg)) return false;

        getUserAgentHeader(&headers);

        string url = strprintf("https://robovinci.xyz/api/submissions/%" PRIu64 "/create", id);

        vector<Http::MimeEntry> query;
        query.emplace_back("file", solutionData, "submission.isl");

        string response;
        if (!Http::postMime(url, query, &headers, &response, pMsg))
        {
            return false;
        }

        return true;
    }
}
