#include "Rects.hpp"
#include "StringUtils.hpp"
#include "Eval.hpp"

using std::string;
using std::vector;

void finishRectImpl(const Problem& problem,
                    const Canvas& initialCanvas,
                    bool last,
                    uint64_t cost,
                    Canvas& canvas,
                    Solution& solution,
                    Canvas& bestCanvas,
                    Solution& bestSolution,
                    uint64_t& bestCost)
{
    auto& blocks = canvas.blocks;
    vector<uint64_t> blockNums;
    for (uint64_t blockNum = 0; blockNum < blocks.size(); blockNum++)
    {
        auto& block = blocks[blockNum];
        if (!block.valid) continue;
        blockNums.push_back(blockNum);
    }

    if (last || blockNums.size() == 1)
    {
        if (cost < bestCost)
        {
            bestCanvas = canvas;
            bestSolution = solution;
            bestCost = cost;
        }

        return;
    }

    string msg;

    uint64_t blockCount = blockNums.size();
    for (uint64_t blockIdx0 = 0; blockIdx0 + 1 < blockCount; blockIdx0++)
    {
        uint64_t blockNum0 = blockNums[blockIdx0];
        Rect rect0 = blocks[blockNum0].rect;
        for (uint64_t blockIdx1 = blockIdx0 + 1; blockIdx1 < blockCount; blockIdx1++)
        {
            uint64_t blockNum1 = blockNums[blockIdx1];
            Rect rect1 = blocks[blockNum1].rect;

            Rect mergeRect(rect0);

            if (rect0.lowX == rect1.lowX &&
                rect0.highX == rect1.highX &&
                rect0.highY == rect1.lowY)
            {
                mergeRect.highY = rect1.highY;
            }
            else if (rect0.lowX == rect1.lowX &&
                     rect0.highX == rect1.highX &&
                     rect0.lowY == rect1.highY)
            {
                mergeRect.lowY = rect1.lowY;
            }
            else if (rect0.lowY == rect1.lowY &&
                     rect0.highY == rect1.highY &&
                     rect0.highX == rect1.lowX)
            {
                mergeRect.highX = rect1.highX;
            }
            else if (rect0.lowY == rect1.lowY &&
                     rect0.highY == rect1.highY &&
                     rect0.lowX == rect1.highX)
            {
                mergeRect.lowX = rect1.lowX;
            }
            else
            {
                continue;
            }

            auto& move = solution.emplace_back();
            move.moveKind = MoveKind::Merge;
            move.blockNums[0] = blockNum0;
            move.blockNums[1] = blockNum1;

            uint64_t curCost;
            if (!evalMove(problem, move, canvas, nullptr, &curCost, nullptr, (uint64_t)-1, nullptr, &msg))
            {
                fprintf(stderr, "%s\n", msg.c_str());
                exit(1);
            }

            finishRectImpl(problem, initialCanvas, last, cost + curCost, canvas, solution, bestCanvas, bestSolution, bestCost);

            blocks.pop_back();
            blocks[blockNum0].valid = true;
            blocks[blockNum1].valid = true;
            canvas.nextBlockId--;

            solution.pop_back();
        }
    }
}

void makeRectImpl(const Problem& problem,
                  const Canvas& initialCanvas,
                  Rect targetRect,
                  bool last,
                  Rect curRect,
                  uint64_t blockNum,
                  uint64_t cost,
                  Canvas& canvas,
                  Solution& solution,
                  Canvas& bestCanvas,
                  Solution& bestSolution,
                  uint64_t& bestCost)
{
    string msg;

    if (curRect == targetRect)
    {
        auto& move = solution.emplace_back();
        move.moveKind = MoveKind::Color;
        move.blockNums[0] = blockNum;
        //move.color = ...;

        uint64_t curCost;
        if (!evalMove(problem, move, canvas, nullptr, &curCost, nullptr, (uint64_t)-1, nullptr, &msg))
        {
            fprintf(stderr, "%s\n", msg.c_str());
            exit(1);
        }

        finishRectImpl(problem, initialCanvas, last, cost + curCost, canvas, solution, bestCanvas, bestSolution, bestCost);

        solution.pop_back();
        return;
    }

    auto& blocks = canvas.blocks;

    if (curRect.lowX != targetRect.lowX &&
        curRect.lowY != targetRect.lowY)
    {
        auto& move = solution.emplace_back();
        move.moveKind = MoveKind::CutXY;
        move.blockNums[0] = blockNum;
        move.x = targetRect.lowX;
        move.y = targetRect.lowY;

        uint64_t curCost;
        if (!evalMove(problem, move, canvas, nullptr, &curCost, nullptr, (uint64_t)-1, nullptr, &msg))
        {
            fprintf(stderr, "%s\n", msg.c_str());
            exit(1);
        }

        uint64_t blockNum2 = blocks.size() - 4 + 2;
        Rect rect2 = blocks[blockNum2].rect;

        makeRectImpl(problem,
                     initialCanvas,
                     targetRect,
                     last,
                     rect2,
                     blockNum2,
                     cost + curCost,
                     canvas,
                     solution,
                     bestCanvas,
                     bestSolution,
                     bestCost);

        blocks.pop_back();
        blocks.pop_back();
        blocks.pop_back();
        blocks.pop_back();
        blocks[blockNum].valid = true;

        solution.pop_back();
    }

    if (curRect.highX != targetRect.highX &&
        curRect.lowY != targetRect.lowY)
    {
        auto& move = solution.emplace_back();
        move.moveKind = MoveKind::CutXY;
        move.blockNums[0] = blockNum;
        move.x = targetRect.highX;
        move.y = targetRect.lowY;

        uint64_t curCost;
        if (!evalMove(problem, move, canvas, nullptr, &curCost, nullptr, (uint64_t)-1, nullptr, &msg))
        {
            fprintf(stderr, "%s\n", msg.c_str());
            exit(1);
        }

        uint64_t blockNum3 = blocks.size() - 4 + 3;
        Rect rect3 = blocks[blockNum3].rect;

        makeRectImpl(problem,
                     initialCanvas,
                     targetRect,
                     last,
                     rect3,
                     blockNum3,
                     cost + curCost,
                     canvas,
                     solution,
                     bestCanvas,
                     bestSolution,
                     bestCost);

        blocks.pop_back();
        blocks.pop_back();
        blocks.pop_back();
        blocks.pop_back();
        blocks[blockNum].valid = true;

        solution.pop_back();
    }

    if (curRect.highX != targetRect.highX &&
        curRect.highY != targetRect.highY)
    {
        auto& move = solution.emplace_back();
        move.moveKind = MoveKind::CutXY;
        move.blockNums[0] = blockNum;
        move.x = targetRect.highX;
        move.y = targetRect.highY;

        uint64_t curCost;
        if (!evalMove(problem, move, canvas, nullptr, &curCost, nullptr, (uint64_t)-1, nullptr, &msg))
        {
            fprintf(stderr, "%s\n", msg.c_str());
            exit(1);
        }

        uint64_t blockNum0 = blocks.size() - 4 + 0;
        Rect rect0 = blocks[blockNum0].rect;

        makeRectImpl(problem,
                     initialCanvas,
                     targetRect,
                     last,
                     rect0,
                     blockNum0,
                     cost + curCost,
                     canvas,
                     solution,
                     bestCanvas,
                     bestSolution,
                     bestCost);

        blocks.pop_back();
        blocks.pop_back();
        blocks.pop_back();
        blocks.pop_back();
        blocks[blockNum].valid = true;

        solution.pop_back();
    }

    if (curRect.lowX != targetRect.lowX &&
        curRect.highY != targetRect.highY)
    {
        auto& move = solution.emplace_back();
        move.moveKind = MoveKind::CutXY;
        move.blockNums[0] = blockNum;
        move.x = targetRect.lowX;
        move.y = targetRect.highY;

        uint64_t curCost;
        if (!evalMove(problem, move, canvas, nullptr, &curCost, nullptr, (uint64_t)-1, nullptr, &msg))
        {
            fprintf(stderr, "%s\n", msg.c_str());
            exit(1);
        }

        uint64_t blockNum1 = blocks.size() - 4 + 1;
        Rect rect1 = blocks[blockNum1].rect;

        makeRectImpl(problem,
                     initialCanvas,
                     targetRect,
                     last,
                     rect1,
                     blockNum1,
                     cost + curCost,
                     canvas,
                     solution,
                     bestCanvas,
                     bestSolution,
                     bestCost);

        blocks.pop_back();
        blocks.pop_back();
        blocks.pop_back();
        blocks.pop_back();
        blocks[blockNum].valid = true;

        solution.pop_back();
    }

    if (curRect.lowX != targetRect.lowX)
    {
        auto& move = solution.emplace_back();
        move.moveKind = MoveKind::CutX;
        move.blockNums[0] = blockNum;
        move.x = targetRect.lowX;

        uint64_t curCost;
        if (!evalMove(problem, move, canvas, nullptr, &curCost, nullptr, (uint64_t)-1, nullptr, &msg))
        {
            fprintf(stderr, "%s\n", msg.c_str());
            exit(1);
        }

        uint64_t blockNum1 = blocks.size() - 2 + 1;
        Rect rect1 = blocks[blockNum1].rect;

        makeRectImpl(problem,
                     initialCanvas,
                     targetRect,
                     last,
                     rect1,
                     blockNum1,
                     cost + curCost,
                     canvas,
                     solution,
                     bestCanvas,
                     bestSolution,
                     bestCost);

        blocks.pop_back();
        blocks.pop_back();
        blocks[blockNum].valid = true;

        solution.pop_back();
    }

    if (curRect.lowY != targetRect.lowY)
    {
        auto& move = solution.emplace_back();
        move.moveKind = MoveKind::CutY;
        move.blockNums[0] = blockNum;
        move.y = targetRect.lowY;

        uint64_t curCost;
        if (!evalMove(problem, move, canvas, nullptr, &curCost, nullptr, (uint64_t)-1, nullptr, &msg))
        {
            fprintf(stderr, "%s\n", msg.c_str());
            exit(1);
        }

        uint64_t blockNum1 = blocks.size() - 2 + 1;
        Rect rect1 = blocks[blockNum1].rect;

        makeRectImpl(problem,
                     initialCanvas,
                     targetRect,
                     last,
                     rect1,
                     blockNum1,
                     cost + curCost,
                     canvas,
                     solution,
                     bestCanvas,
                     bestSolution,
                     bestCost);

        blocks.pop_back();
        blocks.pop_back();
        blocks[blockNum].valid = true;

        solution.pop_back();
    }

    if (curRect.highX != targetRect.highX)
    {
        auto& move = solution.emplace_back();
        move.moveKind = MoveKind::CutX;
        move.blockNums[0] = blockNum;
        move.x = targetRect.highX;

        uint64_t curCost;
        if (!evalMove(problem, move, canvas, nullptr, &curCost, nullptr, (uint64_t)-1, nullptr, &msg))
        {
            fprintf(stderr, "%s\n", msg.c_str());
            exit(1);
        }

        uint64_t blockNum0 = blocks.size() - 2 + 0;
        Rect rect0 = blocks[blockNum0].rect;

        makeRectImpl(problem,
                     initialCanvas,
                     targetRect,
                     last,
                     rect0,
                     blockNum0,
                     cost + curCost,
                     canvas,
                     solution,
                     bestCanvas,
                     bestSolution,
                     bestCost);

        blocks.pop_back();
        blocks.pop_back();
        blocks[blockNum].valid = true;

        solution.pop_back();
    }

    if (curRect.highY != targetRect.highY)
    {
        auto& move = solution.emplace_back();
        move.moveKind = MoveKind::CutY;
        move.blockNums[0] = blockNum;
        move.y = targetRect.highY;

        uint64_t curCost;
        if (!evalMove(problem, move, canvas, nullptr, &curCost, nullptr, (uint64_t)-1, nullptr, &msg))
        {
            fprintf(stderr, "%s\n", msg.c_str());
            exit(1);
        }

        uint64_t blockNum0 = blocks.size() - 2 + 0;
        Rect rect0 = blocks[blockNum0].rect;

        makeRectImpl(problem,
                     initialCanvas,
                     targetRect,
                     last,
                     rect0,
                     blockNum0,
                     cost + curCost,
                     canvas,
                     solution,
                     bestCanvas,
                     bestSolution,
                     bestCost);

        blocks.pop_back();
        blocks.pop_back();
        blocks[blockNum].valid = true;

        solution.pop_back();
    }
}

void makeRect(const Problem& problem,
              Rect rect,
              bool last,
              Canvas& canvas,
              Solution& solution)
{
    uint64_t blockNum = (uint64_t)-1;
    Rect curRect;

    for (uint64_t curBlockNum = 0; curBlockNum < canvas.blocks.size(); curBlockNum++)
    {
        auto& block = canvas.blocks[curBlockNum];
        if (!block.valid) continue;
        if (blockNum != (uint64_t)-1)
        {
            fprintf(stderr, "makeRect: not a single block");
            exit(1);
        }
        blockNum = curBlockNum;
        curRect = block.rect;
    }

    Canvas tmpCanvas(canvas);
    Solution tmpSolution;
    Canvas bestCanvas;
    Solution bestSolution;
    uint64_t bestCost = (uint64_t)-1;

    makeRectImpl(problem, canvas, rect, last, curRect, blockNum, 0, tmpCanvas, tmpSolution, bestCanvas, bestSolution, bestCost);

    if (bestCost == (uint64_t)-1)
    {
        fprintf(stderr, "makeRect: solution not found");
        exit(1);
    }

    canvas = bestCanvas;
    solution.insert(solution.end(), bestSolution.begin(), bestSolution.end());
}

void makeRects(const Problem& problem,
               const vector<Rect>& rects,
               Canvas& canvas,
               Solution& solution)
{
    size_t rectCount = rects.size();
    for (size_t rectIdx = 0; rectIdx < rectCount; rectIdx++)
    {
        Rect rect = rects[rectIdx];
        bool last = (rectIdx + 1 == rectCount);

        makeRect(problem, rect, last, canvas, solution);
    }
}
