#include "Common.hpp"
#include "ProblemCount.hpp"
#include "Best.hpp"
#include "Scores.hpp"

using std::string;
using std::vector;

int main(int argc, char *argv[])
{
    string msg;

    Results results;
    if (!parseResults(&results, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    vector<uint64_t> myScores;
    getMine(results, &myScores);

    vector<uint64_t> bestScores;
    getBest(results, &bestScores);

    uint64_t totalBestScore = 0;
    uint64_t totalSubmittedScore = 0;
    uint64_t totalScore = 0;
    uint64_t totalSimilarity = 0;
    uint64_t totalCost = 0;
    uint64_t totalGap = 0;

    for (uint64_t id = 1; id <= problemCount; id++)
    {
        uint64_t submittedScore = id < myScores.size() ? myScores[id] : (uint64_t)-1;
        uint64_t bestScore = id < bestScores.size() ? bestScores[id] : (uint64_t)-1;

        Best best;
        best.load(id);
        uint64_t score = best.getScore();
        uint64_t similarity = best.getSimilarity();
        uint64_t cost = best.getCost();

        printf("%2" PRIu64 "", id);
        if (bestScore == (uint64_t)-1)
        {
            printf("          ");
        }
        else
        {
            printf("  %8" PRIu64 "", bestScore);
            totalBestScore += bestScore;
        }
        if (submittedScore == (uint64_t)-1)
        {
            printf("          ");
        }
        else
        {
            printf("  %8" PRIu64 "", submittedScore);
            totalSubmittedScore += submittedScore;
        }
        printf("  %8" PRIu64 "  = %8" PRIu64 " + %8" PRIu64 "", score, similarity, cost);

        if (bestScore != (uint64_t)-1)
        {
            uint64_t gap = 0;
            if (score > bestScore)
            {
                gap = score - bestScore;
            }
            printf("  %8" PRIu64 "", gap);
            totalGap += gap;
        }
        printf("\n");

        totalScore += score;
        totalSimilarity += similarity;
        totalCost += cost;
    }

    printf("\n");
    printf("    %8" PRIu64 "  %8" PRIu64 "  %8" PRIu64 "  = %8" PRIu64 " + %8" PRIu64 "  %8" PRIu64 "\n", totalBestScore, totalSubmittedScore, totalScore, totalSimilarity, totalCost, totalGap);

    return 0;
}
