#include "Common.hpp"
#include "ParseUtils.hpp"
#include "Problem.hpp"
#include "ProblemCount.hpp"
#include "Cleanup.hpp"
#include "Best.hpp"
#include "Protocol.hpp"
#include "Scores.hpp"
#include "TimeUtils.hpp"

using std::string;
using std::vector;

void usage()
{
    printf("Usage: submit [<options>] [<id>]\n");
    printf("  <n>   Problem id (default: all)\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
}

int main(int argc, char *argv[])
{
    bool gotId = false;

    bool help = false;
    uint64_t requestedId = 0;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (!gotId)
        {
            if (!parseU64(strArg, &requestedId) || requestedId < 1 || requestedId > 1000)
            {
                usage();
                return 1;
            }
            gotId = true;
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    string msg;

    Protocol::init();
    Cleanup cleanupProtocol([](){ Protocol::cleanup(); });

    string resultsData;
    if (!Protocol::getResults(&resultsData, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    Results results;
    if (!parseResults(resultsData, &results, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    if (!writeResults(resultsData))
    {
        printf("Error writing results\n");
        return 1;
    }

    vector<uint64_t> mine;
    getMine(results, &mine);

    bool submittedAny = false;

    for (uint64_t id = 1; id <= problemCount; id++)
    {
        if (gotId && id != requestedId) continue;

        printf("%2" PRIu64 "  ", id);

        Problem problem;
        if (!loadProblem(id, &problem, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        Best best;
        best.load(id);

        uint64_t score = best.getScore();

        printf("%8" PRIu64 "  ", score);

        uint64_t submittedScore = id < mine.size() ? mine[id] : (uint64_t)-1;

        if (score > submittedScore)
        {
            printf("WORSE than %" PRIu64 "!\n", submittedScore);
            continue;
        }
        else if (score == submittedScore)
        {
            printf("same\n");
            continue;
        }
        fflush(stdout);
        
        string solutionData = formatSolution(problem, best.getSolution());
#if 1
        if (!Protocol::submit(id, solutionData, &msg))
        {
            printf("ERROR!\n");
            printf("%s\n", msg.c_str());
            return 0;
        }
#endif

        submittedAny = true;
        printf("submitted\n");
    }

    if (submittedAny)
    {
        sleepS(2);

        string resultsData;
        if (!Protocol::getResults(&resultsData, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        Results results;
        if (!parseResults(resultsData, &results, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        if (!writeResults(resultsData))
        {
            printf("Error writing results\n");
            return 1;
        }
    }

    return 0;
}
